# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import queue
import asyncio
import grpc
import msgpack


CHANNEL_USABLE_STATE = (grpc.ChannelConnectivity.READY, grpc.ChannelConnectivity.IDLE)


class GRPCClient:

    __slots__ = [
        r'_channel_queue', r'_credentials', r'_options', r'_compression', r'_interceptors',
        r'_request_serializer', r'_response_deserializer',
    ]

    def __init__(
        self, *,
        credentials=None, options=None, compression=None, interceptors=None,
        request_serializer=msgpack.dumps, response_deserializer=msgpack.loads
    ):

        self._channel_queue: queue.SimpleQueue[grpc.aio.Channel] = queue.SimpleQueue()

        self._credentials = credentials
        self._options = options
        self._compression = compression
        self._interceptors = interceptors

        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    async def _make_channel(self, target, timeout) -> grpc.aio.Channel:

        channel = None

        try:

            if self._credentials is None:
                channel = grpc.aio.insecure_channel(
                    target, self._options, self._compression, self._interceptors
                )
            else:
                channel = grpc.aio.secure_channel(
                    target, self._credentials, self._options, self._compression, self._interceptors
                )

            if timeout > 0:
                await asyncio.wait_for(channel.channel_ready(), timeout)
            else:
                await channel.channel_ready()

        except asyncio.TimeoutError as err:

            await channel.close()

            raise err

        setattr(channel, r'name', target)

        return channel

    async def open(self, targets, timeout=0):

        await self.set_channel(targets, timeout)

    async def close(self):

        while not self._channel_queue.empty():
            channel = self._channel_queue.get_nowait()
            await channel.close()

    def get_channel(self) -> grpc.aio.Channel:

        channel = None

        for _ in range(self._channel_queue.qsize()):

            _channel = self._channel_queue.get_nowait()
            self._channel_queue.put_nowait(_channel)

            if _channel.get_state() in CHANNEL_USABLE_STATE:
                channel = _channel
                break

        return channel

    async def set_channel(self, targets, timeout=0):

        channel_dict = {}

        while not self._channel_queue.empty():
            channel = self._channel_queue.get_nowait()
            channel_dict[getattr(channel, r'name')] = channel

        for target in targets:
            if target not in channel_dict:
                channel_dict[target] = await self._make_channel(target, timeout)

        for target, channel in channel_dict.items():
            if target in targets:
                self._channel_queue.put_nowait(channel)
            else:
                await channel.close()

    async def unary_unary(self, method: str, call_params: typing.Dict):

        channel = self.get_channel()

        return await channel.unary_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    def unary_stream(
            self,
            method: str,
            call_params: typing.Dict
    ) -> typing.AsyncIterable:

        channel = self.get_channel()

        return channel.unary_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    async def stream_unary(
            self,
            method: str,
            call_params: typing.Union[typing.Iterable[typing.Dict], typing.AsyncIterable[typing.Dict]]
    ):

        channel = self.get_channel()

        return await channel.stream_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    def stream_stream(
            self,
            method: str,
            call_params: typing.Union[typing.Iterable[typing.Dict], typing.AsyncIterable[typing.Dict]]
    ) -> typing.AsyncIterable:

        channel = self.get_channel()

        return channel.stream_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import json
import yaml

from threading import Lock

from nacos import NacosClient, DEFAULT_GROUP_NAME

from ...extend.base import Utils
from ...extend.config import ConfigureBase


class Configure(ConfigureBase):
    """配置类
    """

    __slots__ = [r'_lock', r'_format', r'_client', r'_listener_callback', r'_listener_params']

    def __init__(
            self, server_addresses, *,
            endpoint=None, namespace=None,
            ak=None, sk=None, username=None, password=None
    ):

        super().__init__()

        self._lock = Lock()
        self._format = None

        self._client = NacosClient(
            server_addresses,
            endpoint=endpoint, namespace=namespace,
            ak=ak, sk=sk, username=username, password=password
        )

        self._listener_params = {}

    def _update_config(self, data=None):

        with self._lock:

            if data is None:
                raw_content = self._client.get_config(**self._listener_params)
            else:
                raw_content = data[r'raw_content']

            self._parser.clear()

            if self._format == r'text':
                self._parser.read_string(raw_content)
            elif self._format == r'json':
                self._parser.read_dict(json.loads(raw_content))
            elif self._format == r'yaml':
                self._parser.read_dict(yaml.load(raw_content, yaml.Loader))

            self._load_options()

        Utils.log.info(f'nacos config update {Utils.md5(raw_content)}')

    def _add_config_watcher(self, data_id, group, timeout, no_snapshot):

        self._listener_params[r'data_id'] = data_id
        self._listener_params[r'group'] = group
        self._listener_params[r'timeout'] = timeout
        self._listener_params[r'no_snapshot'] = no_snapshot

        self._client.add_config_watcher(
            self._listener_params[r'data_id'], self._listener_params[r'group'],
            self._update_config
        )

    def release(self):

        self._client.remove_config_watcher(
            self._listener_params[r'data_id'], self._listener_params[r'group'],
            self._update_config
        )

    def read(self, data_id, group=DEFAULT_GROUP_NAME, timeout=None, no_snapshot=None):

        self._format = r'text'
        self._add_config_watcher(data_id, group, timeout, no_snapshot)
        self._update_config()

    def read_json(self, data_id, group=DEFAULT_GROUP_NAME, timeout=None, no_snapshot=None):

        self._format = r'json'
        self._add_config_watcher(data_id, group, timeout, no_snapshot)
        self._update_config()

    def read_yaml(self, data_id, group=DEFAULT_GROUP_NAME, timeout=None, no_snapshot=None):

        self._format = r'yaml'
        self._add_config_watcher(data_id, group, timeout, no_snapshot)
        self._update_config()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from threading import Lock, Timer

from nacos import NacosClient, DEFAULT_GROUP_NAME
from nacos.listener import SubscribeListener

from ...extend.base import Utils
from ...extend.error import catch_error


class ServiceRegister:
    """服务发现注册
    """

    def __init__(
            self, server_addresses, *,
            endpoint=None, namespace=None,
            ak=None, sk=None, username=None, password=None,
            heartbeat_interval=5
    ):

        super().__init__()

        self._lock = Lock()

        self._client = NacosClient(
            server_addresses,
            endpoint=endpoint, namespace=namespace,
            ak=ak, sk=sk, username=username, password=password
        )

        self._heartbeat_timer = Timer(heartbeat_interval, self._send_heartbeat)
        self._heartbeat_params = {}

    def _send_heartbeat(self):

        with self._lock:
            if self._heartbeat_params:
                self._client.send_heartbeat(**self._heartbeat_params)

    def register(
            self, service_name, ip, port, *,
            cluster_name=None, weight=1.0, metadata=None,
            enable=True, healthy=True, ephemeral=True, group_name=DEFAULT_GROUP_NAME
    ):

        with self._lock:

            self._client.add_naming_instance(
                service_name, ip, port,
                cluster_name=cluster_name, weight=weight, metadata=metadata,
                enable=enable, healthy=healthy, ephemeral=ephemeral, group_name=group_name
            )

            self._heartbeat_params[r'service_name'] = service_name
            self._heartbeat_params[r'ip'] = ip
            self._heartbeat_params[r'port'] = port
            self._heartbeat_params[r'cluster_name'] = cluster_name
            self._heartbeat_params[r'weight'] = weight
            self._heartbeat_params[r'metadata'] = metadata
            self._heartbeat_params[r'ephemeral'] = ephemeral
            self._heartbeat_params[r'group_name'] = group_name

            self._heartbeat_timer.start()

    def deregister(self):

        with self._lock:

            params = {
                r'service_name': self._heartbeat_params[r'service_name'],
                r'ip': self._heartbeat_params[r'ip'],
                r'port': self._heartbeat_params[r'port'],
                r'cluster_name': self._heartbeat_params[r'cluster_name'],
                r'ephemeral': self._heartbeat_params[r'ephemeral'],
                r'group_name': self._heartbeat_params[r'group_name'],
            }

            self._client.remove_naming_instance(**params)

            self._heartbeat_timer.cancel()
            self._heartbeat_params.clear()


class ServiceQuery:
    """服务发现查询
    """

    def __init__(
            self, server_addresses, *,
            endpoint=None, namespace=None,
            ak=None, sk=None, username=None, password=None,
            listener_interval=5
    ):

        super().__init__()

        self._lock = Lock()
        self._hosts = []

        self._client = NacosClient(
            server_addresses,
            endpoint=endpoint, namespace=namespace,
            ak=ak, sk=sk, username=username, password=password
        )

        self._listener_interval = listener_interval
        self._listener_callback = SubscribeListener(self._update_naming_instance, r'listener_callback')
        self._listener_params = {}

    def _update_naming_instance(self, *_):

        with catch_error(), self._lock:

            resp = self._client.list_naming_instance(**self._listener_params)

            if resp[r'hosts']:
                self._hosts = resp[r'hosts']

            Utils.log.info(f'nacos instance update {self._hosts}')

    def start(self, service_name, *, clusters=None, namespace_id=None, group_name=None, healthy_only=True):

        self._listener_params[r'service_name'] = service_name
        self._listener_params[r'clusters'] = clusters
        self._listener_params[r'namespace_id'] = namespace_id
        self._listener_params[r'group_name'] = group_name
        self._listener_params[r'healthy_only'] = healthy_only

        self._client.subscribe(
            self._listener_callback, self._listener_interval,
            **self._listener_params
        )

        self._update_naming_instance()

    def stop(self):

        self._client.unsubscribe(
            self._listener_params[r'service_name']
        )

    def get_host(self):

        return Utils.randhit(self._hosts, lambda x: x[r'weight'])
